#
# MIT License
#
# Copyright (c) 2024 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 1
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. @IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Address Map Finder.
"""

from collections.abc import Callable

import ucdp as u
from ucdp_glbl.finder import find

from .addrmap import AddrMap, Defines
from .addrspaces import Addrspaces

GetAttrspacesFunc = Callable[[], Addrspaces]


def get_addrmap(
    mod: u.BaseMod, defines: Defines | None = None, unique: bool = False, ref: u.TopModRef | None = None
) -> AddrMap:
    """Search Address Spaces and Create Address Map."""
    defines = defines or {}
    addrmap = AddrMap(unique=unique, defines=defines, ref=ref)
    for addrspace in get_addrspaces(mod, defines=defines):
        addrmap.add(addrspace)
    return addrmap


def get_addrspaces(mod: u.BaseMod, defines: Defines | None = None) -> Addrspaces:
    """Search Address Spaces."""
    func = _find_get_addrspaces([mod])
    defines = defines or {}
    return func(**defines)


def _find_get_addrspaces(mods: list[u.BaseMod]) -> GetAttrspacesFunc:
    return find(mods, "get_addrspaces")[1]
